package kjava.net;

import kjava.io.*;
import java.io.*;

public class URL{
  /**
   * The name of the protocol for this URL.
   * The protocol is always stored in lower case.
   */
  private String protocol;

  /**
   * The "authority" portion of the URL.
   */
  private String authority;

  /**
   * The hostname or IP address of this protocol.
   * This includes a possible user. For example <code>joe@some.host.net</code>.
   */
  private String host;

  /**
   * The port number of this protocol or -1 if the port number used is
   * the default for this protocol.
   */
  private int port = -1;	// Initialize for constructor using context.

  /**
   * The "file" portion of the URL. It is defined as <code>path[?query]</code>.
   */
  private String file;

  /**
   * The anchor portion of the URL.
   */
  private String ref;


  public int getPort(){
     return port;
  }

  public String getProtocol(){
     return protocol;
  }

  public URL(String protocol, String host, String file) throws MalformedURLException{
    this(protocol, host, -1, file);
  }

  public URL(String protocol, String host, int port, String file) throws MalformedURLException{
    this.protocol = protocol.toLowerCase();

    this.host = host;
    this.port = port;
    this.authority = null;

    int hashAt = file.indexOf('#');
    if (hashAt < 0){
	this.file = file;
	this.ref = null;
    }else{
	this.file = file.substring(0, hashAt);
	this.ref = file.substring(hashAt + 1);
    }
  }

  public URL(String spec) throws MalformedURLException{
    this((URL) null, spec);
  }

  public URL(URL context, String spec) throws MalformedURLException{
  
    /* A protocol is defined by the doc as the substring before a ':'
     * as long as the ':' occurs before any '/'.
     *
     * If context is null, then spec must be an absolute URL.
     *
     * The relative URL need not specify all the components of a URL.
     * If the protocol, host name, or port number is missing, the value
     * is inherited from the context.  A bare file component is appended
     * to the context's file.  The optional anchor is not inherited. 
     */

    // If this is an absolute URL, then ignore context completely.
    // An absolute URL must have chars prior to "://" but cannot have a colon
    // right after the "://".  The second colon is for an optional port value
    // and implies that the host from the context is used if available.

    int colon;
    if ((colon = spec.indexOf("://", 1)) > 0 &&
	! spec.regionMatches(true, colon, "://:", 0, 4))
      context = null;

    int slash;
    if ((colon = spec.indexOf(':')) > 0 &&
	(colon < (slash = spec.indexOf('/')) || slash < 0))
      {
	// Protocol specified in spec string.
	protocol = spec.substring(0, colon).toLowerCase();
	if (context != null && context.protocol.equals(protocol))
	  {
	    // The 1.2 doc specifically says these are copied to the new URL.
	    host = context.host;
	    port = context.port;
	    file = context.file;
	    authority = context.authority;
	  }
      }
    else if (context != null)
      {
	// Protocol NOT specified in spec string.
	// Use context fields (except ref) as a foundation for relative URLs.
	colon = -1;
	protocol = context.protocol;
	host = context.host;
	port = context.port;
	file = context.file;
        authority = context.authority;
      }
    else	// Protocol NOT specified in spec. and no context available.
      throw new
	  MalformedURLException("Absolute URL required with null context");


    // JDK 1.2 doc for parseURL specifically states that any '#' ref
    // is to be excluded by passing the 'limit' as the indexOf the '#'
    // if one exists, otherwise pass the end of the string.
    int hashAt = spec.indexOf('#', colon + 1);
    parseURL( spec, colon + 1,
			  hashAt < 0 ? spec.length() : hashAt);
    if (hashAt >= 0)
      ref = spec.substring(hashAt + 1);

  }

  protected void parseURL( String spec, int start, int end)
  {
    String host = getHost();
    int port = getPort();
    String file = getFile();
    String ref = getRef();
    
	if (spec.regionMatches (true,start, "//", 0, 2)){
		int hostEnd;
		int colon;

		start += 2;
		int slash = spec.indexOf('/', start);
		if (slash >= 0) 
	  		hostEnd = slash;
        else
	  		hostEnd = end;

		host = spec.substring (start, hostEnd);
	
	// Look for optional port number.  It is valid for the non-port
	// part of the host name to be null (e.g. a URL "http://:80").
	// TBD: JDK 1.2 in this case sets host to null rather than "";
	// this is undocumented and likely an unintended side effect in 1.2
	// so we'll be simple here and stick with "". Note that
	// "http://" or "http:///" produce a "" host in JDK 1.2.
		if ((colon = host.indexOf(':')) >= 0){
	    	try{
				port = Integer.parseInt(host.substring(colon + 1));
	      	}catch (NumberFormatException e){
				; // Ignore invalid port values; port is already set to u's
		  // port.
	      	}
	    	host = host.substring(0, colon);
	  	}
		file = null;
		start = hostEnd;
 	}else if (host == null) 
 		host = "";

	if (file == null || file.length() == 0
		|| (start < end && spec.charAt(start) == '/')) {
	// No file context available; just spec for file.
	// Or this is an absolute path name; ignore any file context.
		file = spec.substring(start+1, end);
		ref = null;
    } 
    else if (start < end)
      {
	// Context is available, but only override it if there is a new file.
        char sepChar = '/';
        int lastSlash = file.lastIndexOf (sepChar);
        if (lastSlash < 0 && File.separatorChar != sepChar
            && getProtocol ().equals ("file"))
          {
            // On Windows, even '\' is allowed in a "file" URL.
            sepChar = File.separatorChar;
            lastSlash = file.lastIndexOf (sepChar);
          }
        
        file = file.substring(0, lastSlash)
                + sepChar + spec.substring (start, end);

        if (getProtocol ().equals ("file"))
          {
            // For "file" URLs constructed relative to a context, we
            // need to canonicalise the file path.
            try
              {
                file = new File (file).getCanonicalPath ();
              }
            catch (IOException e)
              {
              }
          }

	ref = null;
      }

    if (ref == null)
      {
	// Normally there should be no '#' in the file part,
	// but we are nice.
	int hash = file.indexOf('#');
	if (hash != -1)
	  {
	    ref = file.substring(hash + 1, file.length());
	    file = file.substring(0, hash);
	  }
      }

    // XXX - Classpath used to call PlatformHelper.toCanonicalForm() on
    // the file part. It seems like overhead, but supposedly there is some
    // benefit in windows based systems (it also lowercased the string).

    set(getProtocol(), host, port, file, ref);
  }


  protected void set(String protocol, String host, int port, String file,
		     String ref)
  {
    // TBD: Theoretically, a poorly written StreamHandler could pass an
    // invalid protocol.  It will cause the handler to be set to null
    // thus overriding a valid handler.  Callers of this method should
    // be aware of this.
//    this.handler = getURLStreamHandler(protocol);
    this.protocol = protocol.toLowerCase();
    this.authority = null;
    this.port = port;
    this.host = host;
    this.file = file;
    this.ref = ref;
  }


  public String getFile()
  {
    return file == null ? "" : file;
  }

  public String getHost()
  {
    int at = (host == null) ? -1 : host.indexOf('@');
    return at < 0 ? host : host.substring(at + 1, host.length());
  }

  public String getRef()
  {
    return ref;
  }


    public java.io.InputStream openStream() throws java.io.IOException{
        if( getProtocol().equals("file")){
           try{
              return new kjava.io.FileInputStream(getFile());
           }catch (Exception e){
              throw new java.io.IOException(e.toString());
           }
        }else{
           try{
              return javax.microedition.io.Connector.openInputStream(
                              getProtocol() + "://" + getHost() + getFile());
           }catch (Exception e){
              throw new java.io.IOException(e.toString());
           }
        }
    }

  public String toString(){
      return getProtocol() + "://" + getHost() +"/"+ getFile() + 
             (getRef()!=null ? "#" + getRef() : "" );
  }

}

        
